#include <essentials/main.h>
#include <math/main.h>
#include <image/main.h>

#include <cstdio>

using namespace Essentials;
using namespace Math;
using namespace Image;

void Log(const std::string& msg)
{
	printf("%s\n", msg.c_str());
}

void Fourier(const std::string& imageFileName, const std::string& filterFileName, bool invertFilterValues, float filterMinValue, float filterMaxValue)
{
	// load

	byte* data;
	uint16 width, height;
	Image::Format::TYPE format;
	Image::Load(imageFileName, data, width, height, format);
	if (format != Image::Format::R8)
	{
		Log("ERROR: Only grayscale input images are allowed");
		return;
	}

	byte* filterData;
	uint16 filterWidth, filterHeight;
	Image::Format::TYPE filterFormat;
	Image::Load(filterFileName, filterData, filterWidth, filterHeight, filterFormat);
	if (width != filterWidth || height != filterHeight)
	{
		Log("ERROR: Image and filter sizes do not match");
		return;
	}
	if (filterFormat != Image::Format::R8)
	{
		Log("ERROR: Only grayscale filter images are allowed");
		return;
	}

	// fourier

	float* dft = Image::DiscreteFourierTransform_Separable(data, width, height);
	float* dftShifted = (float*)Image::Shift((byte*)dft, width, height, Image::Format::RG32, width / 2, height / 2);

	// apply filter
	if (filterFileName != "")
	{
		// get a and b coefficients of "y = ax + b" rescale function
		Vector2 filterMinMaxRescale_lineCoefficients = Solve2x2(VectorCustom(0.0f, filterMinValue), VectorCustom(1.0f, filterMaxValue));

		for (uint16 j = 0; j < height; j++)
		{
			for (uint16 i = 0; i < width; i++)
			{
				if (invertFilterValues)
					filterData[Idx(i, j, width)] = 255 - filterData[Idx(i, j, width)];
				float filter = (float)filterData[Idx(i, j, width)] / 255.0f;
				filter = filterMinMaxRescale_lineCoefficients.x * filter + filterMinMaxRescale_lineCoefficients.y;

				dftShifted[2*Idx(i, j, width) + 0] *= filter;
				dftShifted[2*Idx(i, j, width) + 1] *= filter;
			}
		}
	}
	delete[] dft;
	dft = (float*)Image::Shift((byte*)dftShifted, width, height, Image::Format::RG32, -width / 2, -height / 2);

	float* idft = InverseDiscreteFourierTransform_Separable(dft, width, height);

	// generate output images

	byte* amplitudesData = new byte[width * height];
	byte* phasesData = new byte[width * height];
	byte* idftRealData = new byte[width * height];

	for (uint16 j = 0; j < height; j++)
	{
		for (uint16 i = 0; i < width; i++)
		{
			float amplitude = Sqrt(Sqr(dftShifted[2*Idx(i, j, width) + 0]) + Sqr(dftShifted[2*Idx(i, j, width) + 1]));
			float phase = ATan2(dftShifted[2*Idx(i, j, width) + 1], dftShifted[2*Idx(i, j, width) + 0]);

			amplitude *= 0.01f;
			phase *= 100.0f;
			phase = fabs(phase);

			amplitudesData[Idx(i, j, width)] = (byte)Clamp(amplitude, 0.0f, 255.0f);
			phasesData[Idx(i, j, width)] = (byte)Clamp(phase, 0.0f, 255.0f);
			idftRealData[Idx(i, j, width)] = (byte)Clamp(idft[2*Idx(i, j, width) + 0], 0.0f, 255.0f);
		}
	}

	// save

	Image::Save(imageFileName + std::string("__amplitudes.png"), amplitudesData, width, height, Image::Format::R8);
	Image::Save(imageFileName + std::string("__phases.png"), phasesData, width, height, Image::Format::R8);
	Image::Save(imageFileName + std::string("__idft_real.png"), idftRealData, width, height, Image::Format::R8);

	// clean

	delete[] dft;
	delete[] dftShifted;
	delete[] idft;
	delete[] amplitudesData;
	delete[] phasesData;
	delete[] idftRealData;
}

int main(int argc, const char* argv[])
{
	if (argc != 6)
	{
		Log("ERROR: Wrong number of input arguments given");
		return 1;
	}

	std::string imageFileName = std::string(argv[1]);
	std::string filterFileName = std::string(argv[2]);
	bool invertFilterValues = atoi(argv[3]) > 0 ? true : false;
	float filterMinValue = (float)atof(argv[4]);
	float filterMaxValue = (float)atof(argv[5]);

	Image::Initialize();
	Image::SetFreeImageCustomOutputMessageFunction(Log);

	Fourier(imageFileName, filterFileName, invertFilterValues, filterMinValue, filterMaxValue);

	Image::Deinitialize();

	return 0;
}
