#include <image/image.h>
#include <image/main.h>
#include <math/common.h>
#include <essentials/macros.h>

#include <FreeImage.h>
#include <squish.h>

#include <cstdio>
#include <cstring>

using namespace Math;

bool Image::Load(const std::string& fileName, byte*& data, uint16& width, uint16& height, Format::TYPE& format)
{
	FREE_IMAGE_FORMAT fif;
	FIBITMAP* fib;

	fif = FreeImage_GetFileType(fileName.c_str());
	if (fif == FIF_UNKNOWN)
		fif = FreeImage_GetFIFFromFilename(fileName.c_str());

	if (fif != FIF_BMP && fif != FIF_JPEG && fif != FIF_PNG)
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't recognize image format: " + fileName);
		return false;
	}

	if (FreeImage_FIFSupportsReading(fif))
		fib = FreeImage_Load(fif, fileName.c_str());

	if (!fib)
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't load image data from file: " + fileName);
		return false;
	}

	byte channelsCount = 0;

	width = FreeImage_GetWidth(fib);
	height = FreeImage_GetHeight(fib);
	if (FreeImage_GetBPP(fib) == 8)
	{
		channelsCount = 1;
		format = Format::R8;
	}
	else if (FreeImage_GetBPP(fib) == 16)
	{
		channelsCount = 2;
		format = Format::RG8;
	}
	else if (FreeImage_GetBPP(fib) == 24)
	{
		channelsCount = 3;
		format = Format::RGB8;
	}
	else if (FreeImage_GetBPP(fib) == 32)
	{
		channelsCount = 4;
		format = Format::RGBA8;
	}
	else
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't recognize image BPP: " + fileName);
		return false;
	}

	uint32 dataSize = width * height * channelsCount;
	data = new byte[dataSize];
	memcpy(data, FreeImage_GetBits(fib), dataSize);

	FreeImage_Unload(fib);

	SAFE_CALL(outputMessageFunction)("OK: Image file loaded: " + fileName);

	return true;
}

bool Image::Save(const std::string& fileName, byte* data, uint16 width, uint16 height, Format::TYPE format)
{
	FREE_IMAGE_FORMAT fif;
	FIBITMAP* fib;

	if (IsCompressed(format) || Is32(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't export data in given format: " + fileName);
		return false;
	}

	fif = FreeImage_GetFIFFromFilename(fileName.c_str());

	if (fif != FIF_BMP && fif != FIF_JPEG && fif != FIF_PNG)
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't recognize file's image format: " + fileName);
		return false;
	}

	fib = FreeImage_Allocate(width, height, BPP(format));

	byte *fiData = FreeImage_GetBits(fib);
	memcpy(fiData, data, Size(width, height, format));

	FreeImage_Save((FREE_IMAGE_FORMAT)fif, fib, fileName.c_str());

	FreeImage_Unload(fib);

	SAFE_CALL(outputMessageFunction)("OK: Image file saved: " + fileName);

	return true;
}

bool Image::IsCompressed(Format::TYPE format)
{
	if (
		format == Format::DXT1 ||
		format == Format::DXT5)
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool Image::Is32(Format::TYPE format)
{
	if (
		format == Format::R32 ||
		format == Format::RG32)
	{
		return true;
	}
	else
	{
		return false;
	}
}

byte Image::BPP(Format::TYPE format)
{
	if (format == Format::R8)
	{
		return 8;
	}
	else if (format == Format::RG8)
	{
		return 16;
	}
	else if (format == Format::RGB8)
	{
		return 24;
	}
	else if (format == Format::RGBA8)
	{
		return 32;
	}
	else if (format == Format::R32)
	{
		return 32;
	}
	else if (format == Format::RG32)
	{
		return 64;
	}
	else
	{
		return 0;
	}
}

byte Image::PixelSize(Format::TYPE format)
{
	if (format == Format::R8)
	{
		return 1;
	}
	else if (format == Format::RG8)
	{
		return 2;
	}
	else if (format == Format::RGB8)
	{
		return 3;
	}
	else if (format == Format::RGBA8)
	{
		return 4;
	}
	else if (format == Format::R32)
	{
		return 4;
	}
	else if (format == Format::RG32)
	{
		return 8;
	}
	else
	{
		return 0;
	}
}

uint32 Image::Size(uint16 width, uint16 height, Format::TYPE format)
{
	if (format == Format::DXT1)
	{
		int blockSize = 8;
		int blocksCount = ( (width + 3)/4 ) * ( (height + 3)/4 );
		return blockSize * blocksCount;
	}
	else if (format == Format::DXT5)
	{
		int blockSize = 16;
		int blocksCount = ( (width + 3)/4 ) * ( (height + 3)/4 );
		return blockSize * blocksCount;
	}
	else
	{
		return PixelSize(format) * width * height;
	}
}

byte Image::MipmapsCount(uint16 width, uint16 height)
{
	byte mipmapsCount = 1;
	uint16 size = Max(width, height);

	if (size == 1)
		return 1;

	do
	{
		size >>= 1;
		mipmapsCount++;
	}
	while (size != 1);

	return mipmapsCount;
}

byte* Image::Scale(byte* data, uint16 width, uint16 height, Format::TYPE& format, uint16 scaledWidth, uint16 scaledHeight, Filter::TYPE filter)
{
	if (IsCompressed(format) || Is32(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't scale data in given format");
		return NULL;
	}

	uint32 scaledDataSize = Size(scaledWidth, scaledHeight, format);
	byte* scaledData = new byte[scaledDataSize];

	FIBITMAP* fibInput = FreeImage_Allocate(width, height, BPP(format));
	memcpy(FreeImage_GetBits(fibInput), data, Size(width, height, format));

	FIBITMAP* fibOutput = FreeImage_Rescale(fibInput, scaledWidth, scaledHeight, (FREE_IMAGE_FILTER)filter);
	memcpy(scaledData, FreeImage_GetBits(fibOutput), scaledDataSize);

	FreeImage_Unload(fibInput);
	FreeImage_Unload(fibOutput);

	return scaledData;
}

std::vector<Image::Mipmap> Image::GenerateMipmaps(byte* data, uint16 width, uint16 height, Format::TYPE format, Filter::TYPE filter)
{
	std::vector<Mipmap> mipmaps;

	if (IsCompressed(format) || Is32(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't generate mipmaps in given format");
		return mipmaps;
	}
	if (!IsPowerOfTwo(width) || !IsPowerOfTwo(height))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't generate mipmaps for non-power-of-two image");
		return mipmaps;
	}

	byte mipmapsCount = MipmapsCount(width, height);
	mipmaps.resize(mipmapsCount);
	mipmaps[0].data = new byte[Size(width, height, format)];

	for (byte i = 0; i < mipmapsCount; i++)
	{
		mipmaps[i].width = width;
		mipmaps[i].height = height;

		if (width > 1)
			width >>= 1;
		if (height > 1)
			height >>= 1;
	}

	memcpy(mipmaps[0].data, data, Size(mipmaps[0].width, mipmaps[0].height, format));

	for (byte i = 1; i < mipmaps.size(); i++)
		mipmaps[i].data = Scale(mipmaps[i-1].data, mipmaps[i-1].width, mipmaps[i-1].height, format, mipmaps[i].width, mipmaps[i].height, filter);

	return mipmaps;
}

void Image::SwapChannels(byte* data, uint16 width, uint16 height, Format::TYPE format, byte firstChannelIndex, byte secondChannelIndex)
{
	if (IsCompressed(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't swap channels in given format");
		return;
	}

	byte pixelSize = PixelSize(format);
	uint32 pixelsCount = Size(width, height, format) / pixelSize;

	for (uint32 i = 0; i < pixelsCount; i++)
	{
		byte temp = data[i*pixelSize + firstChannelIndex];
		data[i*pixelSize + firstChannelIndex] = data[i*pixelSize + secondChannelIndex];
		data[i*pixelSize + secondChannelIndex] = temp;
	}
}

byte* Image::Shift(byte* data, uint16 width, uint16 height, Format::TYPE format, uint16 xOffset, uint16 yOffset)
{
	if (IsCompressed(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't shift data in given format");
		return NULL;
	}

	// normalize offsets

	if (xOffset >= 0)
	{
		xOffset = xOffset % width;
	}
	else
	{
		while (xOffset < 0)
			xOffset += width;
	}

	if (yOffset >= 0)
	{
		yOffset = yOffset % height;
	}
	else
	{
		while (yOffset < 0)
			yOffset += height;
	}

	// shift

	byte pixelSize = PixelSize(format);
	uint32 dataSize = Size(width, height, format);
	byte* shiftedData = new byte[dataSize];

	for (uint16 j = 0; j < height; j++)
	{
		for (uint16 i = 0; i < width; i++)
		{
			int x = (i + xOffset) % width;
			int y = (j + yOffset) % height;

			byte* srcData = &data[pixelSize * Idx(i, j, width)];
			byte* dstData = &shiftedData[pixelSize * Idx(x, y, width)];

			memcpy(dstData, srcData, pixelSize);
		}
	}

	//

	return shiftedData;
}

byte* Image::Compress(byte* data, uint16 width, uint16 height, Format::TYPE desiredFormat)
{
	if (!IsCompressed(desiredFormat))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't compress to desired format");
		return NULL;
	}

	int flags = 0;

	if (desiredFormat == Format::DXT1)
		flags = squish::kDxt1;
	else if (desiredFormat == Format::DXT5)
		flags = squish::kDxt5;

	uint32 dataSize = Size(width, height, Format::RGBA8);
	uint32 compressedDataSize = squish::GetStorageRequirements(width, height, flags);
	byte* compressedData = new byte[compressedDataSize];

	squish::CompressImage(data, width, height, compressedData, flags);

	return compressedData;
}

byte* Image::Decompress(byte* data, uint16 width, uint16 height, Format::TYPE format)
{
	if (!IsCompressed(format))
	{
		SAFE_CALL(outputMessageFunction)("ERROR: Can't decompress data in given format");
		return NULL;
	}

	int flags = 0;

	if (format == Format::DXT1)
		flags = squish::kDxt1;
	else if (format == Format::DXT5)
		flags = squish::kDxt5;

	uint32 dataSize = Size(width, height, Format::RGBA8);
	uint32 compressedDataSize = squish::GetStorageRequirements(width, height, flags);
	byte* decompressedData = new byte[dataSize];

	squish::DecompressImage(decompressedData, width, height, data, flags);

	return decompressedData;
}
